@@ -5,7 +5,7 @@ import glob
 from pathlib import Path
 from fnmatch import fnmatch
 from types import SimpleNamespace
-from typing import List, Tuple, Optional
+from typing import List, Tuple, Optional, Union
 import json
 import argparse
 from argparse import Namespace
@@ -14,6 +14,7 @@ import concurrent.futures
 import queue
 import threading
 from itertools import chain
+from collections import defaultdict
 
 import xml.etree.ElementTree as ET
 import dask.array as da
@@ -23,6 +24,13 @@ import numpy as np
 from . import blur
 from . import __version__
 
+import logging
+logging.basicConfig(
+	format="{asctime} - {levelname} - {message}",
+	style="{",
+	datefmt="%Y-%m-%d %H:%M",
+)
+
 def center_crop(A, shape):
 	"""Crop numpy array to (h, w) from center."""
 	h, w = shape[-2:]
@@ -36,6 +44,7 @@ def load_flats(flat_field_tag, shape=None, **kwargs):
 	files = sorted(glob.glob(flat_field_tag + '*'))
 	for file in files:
 		im = np.load(file)['im']
+		assert all(a >= b for a, b in zip(im.shape, shape)), 'flat file is smaller than image'
 		if shape is not None:
 			im = center_crop(im, shape)
 		cim = cp.array(im,dtype=cp.float32)
@@ -87,6 +96,17 @@ class Container:
 			arr = arr.astype(dtype)
 		return arr
 
+	def __eq__(self, other):
+		if isinstance(other, Container):
+			return self.path == other.path
+		elif isinstance(other, str):
+			return self.path == other
+		return False
+
+	def __hash__(self):
+		return hash(self.path)
+
+
 	def compute(self):
 		"""Compute only the current container (recursively if channels)."""
 		if isinstance(self.data, list):
@@ -178,18 +198,43 @@ class FolderFilter:
 		self.start_parts = self._parse_pattern(self.start_pattern)
 		self.end_parts = self._parse_pattern(self.end_pattern)
 		
-	def _parse_pattern(self, pattern: str) -> Optional[Tuple]:
+	def _parse_pattern(self, haystack: str) -> Optional[Tuple]:
 		"""Parse a pattern using the regex to extract components"""
-		match = self.regex.match(pattern)
+		match = self.regex.match(haystack)
 		if match:
 			return match.groups()
 		return None
-	
-	def _extract_numeric_part(self, text: str) -> int:
-		"""Extract numeric part from text like 'H1' -> 1"""
-		match = re.search(r'\d+', text)
-		return int(match.group()) if match else 0
-	
+
+	def _undo_regex(self, groups: Union[Tuple, re.Match, None]) -> Optional[str]:
+		"""
+		Reconstruct a string from captured groups based on the stored regex pattern.
+		If `groups` is a re.Match it will use match.groups().
+		If you pass fewer groups than the regex has, reconstruction stops right
+		after the last provided group (no extra literal text appended).
+		"""
+		if groups is None:
+			return None
+
+		# Accept either a tuple-of-groups or a Match object.
+		if hasattr(groups, "groups"):
+			groups = groups.groups()
+		groups = tuple(groups)
+
+		pattern = self.regex.pattern
+		parts = re.split(r'\([^)]*\)', pattern)  # literal chunks between capture groups
+
+		rebuilt = []
+		for i, g in enumerate(groups):
+			literal = parts[i] if i < len(parts) else ''
+			rebuilt.append(literal)
+			rebuilt.append('' if g is None else str(g))
+
+		# Append the trailing literal only if all groups were provided.
+		if len(groups) == len(parts) - 1:
+			rebuilt.append(parts[len(groups)])
+
+		return ''.join(rebuilt)
+
 	def _compare_patterns(self, file_parts: Tuple, start_parts: Tuple, end_parts: Tuple) -> bool:
 		"""
 		Compare if file_parts falls within the range defined by start_parts and end_parts
@@ -233,16 +278,6 @@ class FolderFilter:
 			return False
 		return self._compare_patterns(file_parts, self.start_parts, self.end_parts)
 	
-	def filter_files(self, filenames: List[str]) -> List[str]:
-		"""Filter filenames that fall within the specified range"""
-		matching_files = []
-		
-		for filename in filenames:
-			if self.isin(filename):
-				matching_files.append(filename)
-				
-		return matching_files
-
 	def get_matches(self, folders):
 		matches = dict()
 		for root in folders:
@@ -271,16 +306,23 @@ class FolderFilter:
 class Block(list):
 	def __init__(self, items=None):
 		self.background = None
-		if isinstance(items, (list, tuple)):
-			for item in items:
-				self.append(item)
-		elif items:
-			self.append(items)
+	def add(self, image):
+		fov, tag = path_parts(image.path)
+		if self.fov() == fov or not self:
+			if self:
+				del self[-1].data
+			self.append(image)
+			return True
+		else:
+			return False
+	
 	def parts(self):
 		path = self[0].path
 		fov, tag = path_parts(path)
 		return fov, tag
 	def fov(self):
+		if not self:
+			return None
 		fov,_ = self.parts()
 		return fov
 	def tag(self):
@@ -288,10 +330,13 @@ class Block(list):
 		return tag
 	def iset(self):
 		return get_iset(self[0].path)
+	def __repr__(self):
+		paths = [image.path for image in self]
+		return f"Block({paths})"
 	
 class ImageQueue:
 	__version__ = __version__
-	def __init__(self, args, prefetch_count=6):
+	def __init__(self, args, prefetch_count=5):
 		self.args = args
 		self.args_array = namespace_to_array(self.args.settings)
 		self.__dict__.update(vars(args.paths))
@@ -301,12 +346,44 @@ class ImageQueue:
 		fov_min, fov_max = (-float('inf'), float('inf'))
 		if hasattr(self, "fov_range"):
 			fov_min, fov_max = map(float, self.fov_range.split(':'))
-		matches = FolderFilter(self.hyb_range, self.regex, fov_min, fov_max).get_matches(self.hyb_folders)
+		
+		self.background_files = set()
 		background = None
 		if hasattr(self, "background_range"):
-			background = FolderFilter(self.background_range, self.regex, fov_min, fov_max).get_matches(self.hyb_folders)
-			self.background = True
+			filtered = FolderFilter(self.background_range, self.regex, fov_min, fov_max)
+			background = filtered.get_matches(self.hyb_folders)
+			self.background_files = set(chain.from_iterable(background.values()))
+		filtered = FolderFilter(self.hyb_range, self.regex, fov_min, fov_max)
+		matches = filtered.get_matches(self.hyb_folders)
 
+		# do output summary, eventually replace with terminal gui
+		counts = defaultdict(lambda: defaultdict(int))
+		rounds = dict()
+		for iset,ifov in matches:
+			for path in matches[(iset,ifov)]:
+				base = os.path.basename(os.path.dirname(path))
+				parts = list(filtered._parse_pattern(base))
+				parts = parts[:3]
+				undone = filtered._undo_regex(parts)
+				counts[iset][undone] += 1
+				rounds[undone] = True
+		rounds = sorted(rounds)
+		width = max(len(r) for r in rounds) + 1
+		summary = ''
+		summary += 'SUMMARY'.center(len(rounds) * width + width, '-')	
+		summary += '\n'
+		summary += 'set'.ljust(width)
+		for undone in rounds:
+			summary += undone.rjust(width)
+		summary += '\n'
+		for iset in counts:
+			summary += str(iset).ljust(width)
+			for undone in rounds:
+				summary += str(counts[iset][undone]).rjust(width)
+			summary += '\n'
+		summary += ''.center(len(rounds) * width + width, '-')
+		self.summary = summary
+			
 		# Peek at first image to set shape/dtype
 		first_image = None
 		for path in chain.from_iterable(matches.values()):
@@ -322,18 +399,16 @@ class ImageQueue:
 	
 		# interlace the background with the regular images
 		shared = set(matches.keys()).intersection(background.keys()) if background else matches.keys()
-		#matches = [item for key in shared for item in matches[key]]
-		#background = [item for key in shared for item in background[key]] if background else None
+
 		interlaced = []
 		for key in shared:
 			if background and key in background:
 				interlaced.extend(background[key])  # put background first
 			hsorted = self.hsorted(matches[key])
-			interlaced.extend(hsorted)			    # then all matches for that iset,ifov
+			interlaced.extend(hsorted)				# then all matches for that iset,ifov
 
 		self.files = iter(interlaced)
 
-		self.block = Block()
 		# Start worker thread(s)
 		self.queue = queue.Queue(maxsize=prefetch_count)
 		self.stop_event = threading.Event()
@@ -351,13 +426,17 @@ class ImageQueue:
 		fit_files = list()
 		# check if the fov has been fitted
 		fov, tag = path_parts(path)
+
+		# we should test whether the load worked!!!!!!!!!!
 		for icol in range(self.shape[0] - 1):
 			filename = self.hyb_save.format(fov=fov, tag=tag, icol=icol)
 			filepath = os.path.join(self.output_folder, filename)
-			if not os.path.exists(filepath):
+			# only compute if the image has not been processed or if it is background
+			if not os.path.exists(filepath) or container in self.background_files:
 				container[icol].compute()
 			else:
-				container[icol].fits = cp.load(filepath)
+				Xh = cp.load(filepath)['Xh']
+				setattr(container, f'col{icol}', Xh)
 		icol += 1
 		filename = self.dapi_save.format(fov=fov, tag=tag, icol=icol)
 		filepath = os.path.join(self.output_folder, filename)
@@ -366,6 +445,16 @@ class ImageQueue:
 		else:
 			container.Xh_plus = cp.load(filepath)['Xh_plus']
 			container.Xh_minus = cp.load(filepath)['Xh_minus']
+		# attach drift file to container
+		iset = get_iset(path)
+		filename = self.drift_save.format(fov=fov, iset=iset)
+		filepath = os.path.join(self.output_folder, filename)
+		if os.path.exists(filepath):
+			drifts, files, ref, ref_path = cp.load(filepath, allow_pickle=True)
+			container.drifts = drifts
+			container.drift_files = files
+			container.drift_ref = ref
+			container.drift_ref_path = ref_path
 		return container
 
 	def _worker(self):
@@ -381,130 +470,24 @@ class ImageQueue:
 				#dummy = lambda : None
 				#dummy.path = path
 				#self.queue.put(dummy)
-				self.queue.put(False)
+				#self.queue.put(False)
 				continue
 		# Signal no more images
+		container.last = True
 		self.queue.put(None)
 
 	def __iter__(self):
 		return self
 
-	'''
 	def __next__(self):
 		img = self.queue.get()
 		if img is None:
 			raise StopIteration
 		return img
-	'''
 	
-	def __next__(self):
-		"""Return the next block of images (same FOV)."""
-		block = self.block
-		first_item = self.queue.get()
-
-		if first_item is None:
-			if block:
-				self.block = Block()
-				self.queue.put(None)
-				return block
-			else:
-				raise StopIteration
-		
-		if hasattr(self, "background_range"):
-			self.block.background = first_item
-		else:
-			block.append(first_item)
-
-		#ifov = None if first_item == False else get_ifov(first_item.path)
-		#iset = None if first_item == False else get_iset(first_item.path)
-		ifov = get_ifov(first_item.path) if first_item else first_item
-		iset = get_iset(first_item.path) if first_item else first_item
-		
-		# Keep consuming queue until FOV changes or None
-		while True:
-			item = self.queue.get()
-			if item == False:
-				break
-			if item is None:
-				self.queue.put(None)
-				self.block = Block()
-				break
-			if (get_ifov(item.path) != ifov) or (get_iset(item.path) != iset):
-				if hasattr(self, "background_range"):
-					self.block.background = item
-				else:
-					self.block = Block(item)
-				break
-			block.append(item)
-		if hasattr(self, "background_range") and first_item == False:
-			block.clear()
-		block.ifov = ifov
-		return block
-
-	'''
-	def __next__(self):
-		"""Return the next block of images (same FOV)."""
-		logger.debug(f"__next__ called, queue size: {getattr(self.queue, 'qsize', lambda: 'unknown')()}")
-		
-		block = Block()  # Fresh block each time
-		
-		# Get first item with debugging
-		logger.debug("Getting first item from queue...")
-		start_time = time.time()
-		first_item = self.queue.get()
-		get_time = time.time() - start_time
-		logger.debug(f"Got first item in {get_time:.3f}s: {type(first_item).__name__}")
-		
-		if first_item is None:
-			logger.debug("First item is None, raising StopIteration")
-			raise StopIteration
-		
-		block.append(first_item)
-		
-		# Handle FOV logic with debugging
-		if first_item == False:
-			ifov = None
-			logger.debug("First item is False, ifov = None")
-		else:
-			logger.debug(f"Getting ifov for: {getattr(first_item, 'path', 'NO_PATH_ATTR')}")
-			ifov = get_ifov(first_item.path)
-			logger.debug(f"ifov = {ifov}")
-		
-		# Keep consuming queue until FOV changes or None
-		item_count = 1
-		while True:
-			logger.debug(f"Loop iteration {item_count}, getting next item...")
-			start_time = time.time()
-			item = self.queue.get()
-			get_time = time.time() - start_time
-			logger.debug(f"Got item {item_count} in {get_time:.3f}s: {type(item).__name__}")
-			
-			if item == False:
-				logger.debug("Got False, breaking")
-				break
-			if item is None:
-				logger.debug("Got None, putting back and breaking")
-				self.queue.put(None)
-				break
-			
-			logger.debug(f"Getting ifov for item {item_count}: {getattr(item, 'path', 'NO_PATH_ATTR')}")
-			item_ifov = get_ifov(item.path)
-			logger.debug(f"Item ifov: {item_ifov}, current ifov: {ifov}")
-			
-			if item_ifov != ifov:
-				logger.debug("FOV changed, putting item back and breaking")
-				self.queue.put(item)  # Put back the item with different FOV
-				break
-			
-			block.append(item)
-			item_count += 1
-			logger.debug(f"Added item to block, block size now: {len(block)}")
-		
-		logger.debug(f"Returning block with {len(block)} items")
-		return block
-	'''
-
 	'''
+		# this code is so we can eventually add a --watch parameter so mermake keeps
+		# runnning after processing all fovs and then processes any new fovs that appear
 		# If we reach here, there are no more images in the current batch
 		if False:
 			# In watch mode, look for new files
@@ -542,9 +525,26 @@ class ImageQueue:
 	def __exit__(self, exc_type, exc_val, exc_tb):
 		self.close()
 
+	'''
 	def close(self):
 		self.stop_event.set()
 		self.thread.join()
+	'''
+	def close(self):
+		self.stop_event.set()
+    
+		# Drain the queue so the worker thread can finish
+		try:
+			while True:
+				self.queue.get_nowait()
+		except queue.Empty:
+			pass
+    
+		# Use a timeout to avoid hanging indefinitely
+		self.thread.join(timeout=2.0)
+    
+		if self.thread.is_alive():
+			print("Warning: worker thread did not terminate", flush=True)
 
 	def _is_fitted(self, path):
 		fov, tag = path_parts(path)
@@ -559,20 +559,25 @@ class ImageQueue:
 			return False
 		return True
 
-	def save_hyb(self, path, icol, Xhf, attempt=1, max_attempts=3):
-		fov,tag = path_parts(path)
-		filename = self.hyb_save.format(fov=fov, tag=tag, icol=icol)
-		filepath = os.path.join(self.output_folder, filename)
-	
-		Xhf = [x for x in Xhf if x.shape[0] > 0]
-		if Xhf:
-			xp = cp.get_array_module(Xhf[0])
-			Xhf = xp.vstack(Xhf)
+	def save_xfits(self, image, icol=-1, attempt=1, max_attempts=3):
+		fov, tag = path_parts(image.path)
+		
+		# determine filename and data based on icol
+		if icol == -1:
+			filename = self.dapi_save.format(fov=fov, tag=tag)
+			data = {
+				'Xh_plus': image.Xh_plus,
+				'Xh_minus': image.Xh_minus,
+			}
 		else:
-			xp = np
-			Xhf = np.array([])
-		if not os.path.exists(filepath) or (hasattr(self, "redo") and self.redo):
-			xp.savez_compressed(filepath, Xh=Xhf, version=__version__, args=self.args_array)
+			filename = self.hyb_save.format(fov=fov, tag=tag, icol=icol)
+			Xh = getattr(image, f'col{icol}')
+			data = {'Xh': Xh}
+		
+		# save if file doesn't exist
+		filepath = os.path.join(self.output_folder, filename)
+		if not os.path.exists(filepath):
+			cp.savez_compressed(filepath, version=__version__, args=self.args_array, **data)
 			#  Optional integrity check after saving
 			# this seems to greatly slow everything down
 			#try:
@@ -584,32 +589,6 @@ class ImageQueue:
 			#		return self.save_hyb(path, icol, Xhf, attempt=attempt+1, max_attempts=max_attempts)
 			#	else:
 			#		raise RuntimeError(f"Failed saving xfit file after {max_attempts} attempts: {filepath}")
-		del Xhf
-		if xp == cp:
-			xp._default_memory_pool.free_all_blocks()  # Free standard GPU memory pool
-
-	def save_dapi(self, path, icol, Xh_plus, Xh_minus, attempt=1, max_attempts=3):
-		fov, tag = path_parts(path)
-		filename = self.dapi_save.format(fov=fov, tag=tag, icol=icol)
-		filepath = os.path.join(self.output_folder, filename)
-	
-		xp = cp.get_array_module(Xh_plus)
-		if not os.path.exists(filepath) or (hasattr(self, "redo") and self.redo):
-			xp.savez_compressed(filepath, Xh_plus=Xh_plus, Xh_minus=Xh_minus, version=__version__, args=self.args_array)
-			#  Optional integrity check after saving
-			# this seems to greatly slow everything down
-			#try:
-			#	with np.load(filepath) as dat:
-			#		_ = dat["Xh_minus"].shape  # Try accessing a key
-			#except Exception as e:
-			#	os.remove(filepath)
-			#	if attempt < max_attempts:
-			#		return self.save_dapi(path, icol, Xh_plus, Xh_minus, attempt=attempt+1, max_attempts=max_attempts)
-			#	else:
-			#		raise RuntimeError(f"Failed saving xfit file after {max_attempts} attempts: {filepath}")
-		del Xh_plus, Xh_minus
-		if xp == cp:
-			xp._default_memory_pool.free_all_blocks()
 
 def read_xml(path):
 	# Open and parse the XML file
